#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`bankid.certutils` -- Certificate Utilities
================================================

Created 2016-04-10 by hbldh

"""

import os
import subprocess

from bankid.certs import get_test_cert_p12
from bankid.exceptions import BankIDError

_TEST_CERT_PASSWORD = "qwerty123"
_TEST_CERT_URL = "https://www.bankid.com/assets/bankid/rp/FPTestcert4_20220818.p12"


def create_bankid_test_server_cert_and_key(destination_path):
    """Split the bundled test certificate into certificate and key parts and save them
    as separate files, stored in PEM format.

    If the environment variable TEST_CERT_FILE is set, use this file
    instead of fetching the P12 certificate.

    :param destination_path: The directory to save certificate and key files to.
    :type destination_path: str
    :returns: The path tuple ``(cert_path, key_path)``.
    :rtype: tuple

    """
    if os.getenv("TEST_CERT_FILE"):
        certificate, key = split_certificate(
            os.getenv("TEST_CERT_FILE"), destination_path, password=_TEST_CERT_PASSWORD
        )

    else:
        # Fetch testP12 certificate path
        certificate, key = split_certificate(str(get_test_cert_p12()), destination_path, password=_TEST_CERT_PASSWORD)

    # Return path tuples.
    return certificate, key


def split_certificate(certificate_path, destination_folder, password=None):
    """Splits a PKCS12 certificate into Base64-encoded DER certificate and key.

    This method splits a potentially password-protected
    `PKCS12 <https://en.wikipedia.org/wiki/PKCS_12>`_ certificate
    (format ``.p12`` or ``.pfx``) into one certificate and one key part, both in
    `pem <https://en.wikipedia.org/wiki/X.509#Certificate_filename_extensions>`_
    format.

    :returns: Tuple of certificate and key string data.
    :rtype: tuple

    """
    try:
        # Attempt Linux and Darwin call first.
        p = subprocess.Popen(["openssl", "version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        sout, serr = p.communicate()
        openssl_executable_version = sout.decode().lower()
        if not (openssl_executable_version.startswith("openssl") or openssl_executable_version.startswith("libressl")):
            raise BankIDError("OpenSSL executable could not be found. " "Splitting cannot be performed.")
        openssl_executable = "openssl"
    except Exception:
        # Attempt to call on standard Git for Windows path.
        p = subprocess.Popen(
            ["C:\\Program Files\\Git\\mingw64\\bin\\openssl.exe", "version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        sout, serr = p.communicate()
        if not sout.decode().lower().startswith("openssl"):
            raise BankIDError("OpenSSL executable could not be found. " "Splitting cannot be performed.")
        openssl_executable = "C:\\Program Files\\Git\\mingw64\\bin\\openssl.exe"

    if not os.path.exists(os.path.abspath(os.path.expanduser(destination_folder))):
        os.makedirs(os.path.abspath(os.path.expanduser(destination_folder)))

    # Paths to output files.
    out_cert_path = os.path.join(os.path.abspath(os.path.expanduser(destination_folder)), "certificate.pem")
    out_key_path = os.path.join(os.path.abspath(os.path.expanduser(destination_folder)), "key.pem")

    # Use openssl for converting to pem format.
    pipeline_1 = [
        openssl_executable,
        "pkcs12",
        "-in",
        "{0}".format(certificate_path),
        "-passin" if password is not None else "",
        "pass:{0}".format(password) if password is not None else "",
        "-out",
        "{0}".format(out_cert_path),
        "-clcerts",
        "-nokeys",
    ]
    p = subprocess.Popen(list(filter(None, pipeline_1)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()

    if p.returncode:
        raise BankIDError("Error converting certificate: {0}".format(err.decode("utf-8")))

    pipeline_2 = [
        openssl_executable,
        "pkcs12",
        "-in",
        "{0}".format(certificate_path),
        "-passin" if password is not None else "",
        "pass:{0}".format(password) if password is not None else "",
        "-out",
        "{0}".format(out_key_path),
        "-nocerts",
        "-nodes",
    ]
    p = subprocess.Popen(list(filter(None, pipeline_2)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()

    if p.returncode:
        raise BankIDError("Error converting certificate: {0}".format(err.decode("utf-8")))

    # Return path tuples.
    return out_cert_path, out_key_path


def main(verbose=True):
    paths = create_bankid_test_server_cert_and_key(os.path.expanduser("~"))
    if verbose:
        print("Saved certificate as {0}".format(paths[0]))
        print("Saved key as {0}".format(paths[1]))
    return paths


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import requests

import bankid
from bankid.certs import get_test_cert_and_key


@pytest.fixture(scope="module")
def ip_address():
    return requests.get("https://httpbin.org/ip").json()["origin"].split(",")[0]


@pytest.fixture(scope="session")
def cert_and_key(tmpdir_factory):
    cert, key = get_test_cert_and_key()
    return cert, key



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`bankid.exceptions` -- PyBankID Exceptions
===============================================

.. moduleauthor:: hbldh <henrik.blidh@nedomkull.com>

Created on 2014-09-10, 08:29

"""


def get_json_error_class(response):
    data = response.json()
    error_class = _JSON_ERROR_CODE_TO_CLASS.get(data.get("errorCode"), BankIDError)
    return error_class("{0}: {1}".format(data.get("errorCode"), data.get("details")))


class BankIDError(Exception):
    """Parent exception class for all PyBankID errors."""

    def __init__(self, *args, **kwargs):
        super(BankIDError, self).__init__(*args, **kwargs)
        self.rfa = None


class BankIDWarning(Warning):
    """Warning class for PyBankID."""

    pass


class InvalidParametersError(BankIDError):
    """User induced error.

    **Code:** ``invalidParameters``

    **Reason:** Invalid parameter. Invalid use of method.

    **Action by RP:** RP must not try the same request again.
    This is an internal error within RP's system and must not be '
    communicated to the user as a BankID error.

    """


class AlreadyInProgressError(BankIDError):
    """Failure to create new order due to one already in progress.

    **Code:** ``alreadyInProgress``

    **Reason:** An order for this user is already in progress. The order
    is aborted. No order is created.

    **Action by RP:** RP must inform the user that a login or signing
    operation is already initiated for this user.
    Message RFA4 should be used.

    """

    def __init__(self, *args, **kwargs):
        super(AlreadyInProgressError, self).__init__(*args, **kwargs)
        self.rfa = 4


class InternalError(BankIDError):
    """Remote server error.

    **Code:** ``internalError``

    **Reason:** Internal technical error in the BankID system.

    **Action by RP:** RP must not automatically try again. RP must
    inform the user that a technical error has
    occurred. Message RFA5 should be used.

    """

    def __init__(self, *args, **kwargs):
        super(InternalError, self).__init__(*args, **kwargs)
        self.rfa = 5


class MaintenanceError(BankIDError):
    """The service is temporarily out of service.

    **Code:** ``maintenance``

    **Action by RP:** RP may try again without informing the user.
    If this error is returned repeatedly, RP must inform the user.
    Message RFA5.

    """

    def __init__(self, *args, **kwargs):
        super(MaintenanceError, self).__init__(*args, **kwargs)
        self.rfa = 5


class UnauthorizedError(BankIDError):
    """RP does not have access to the service.

    **Code:** ``unauthorized``

    **Action by RP:** RP must not try the same request again.
    This is an internal error within RP's system and must not be '
    communicated to the user as a BankID error.

    """

    pass


class NotFoundError(BankIDError):
    """An erroneously URL path was used.

    **Code:** ``notFound``

    **Action by RP:** RP must not try the same request again.
    This is an internal error within RP's system and must not be '
    communicated to the user as a BankID error.

    """

    pass


class RequestTimeoutError(BankIDError):
    """It took too long time to transmit the request.

    **Code:** ``requestTimeout``

    **Action by RP:** RP must not automatically try again.
    This error may occur if the processing at RP or the communication is too
    slow. RP must inform the user. Message RFA5

    """

    def __init__(self, *args, **kwargs):
        super(RequestTimeoutError, self).__init__(*args, **kwargs)
        self.rfa = 5


_JSON_ERROR_CODE_TO_CLASS = {
    "invalidParameters": InvalidParametersError,
    "alreadyInProgress": AlreadyInProgressError,
    "unauthorized": UnauthorizedError,
    "notFound": NotFoundError,
    # 'methodNotAllowed': ,  # This will not be handled here...
    "requestTimeout": RequestTimeoutError,
    # 'unsupportedMediaType': ,  # This will not be handled here...
    "internalError": InternalError,
    "maintenance": MaintenanceError,
}



# This directory contains the root certificates used by BankID.
# We have to pin these to prevent basic MITM attacks.

from pathlib import Path


def get_test_cert_p12():
    return (Path(__file__).parent / "FPTestcert4_20220818.p12").resolve()


def get_test_cert_and_key():
    return (
        (Path(__file__).parent / "FPTestcert4_20220818_cert.pem").resolve(),
        (Path(__file__).parent / "FPTestcert4_20220818_key.pem").resolve(),
    )



import base64
import datetime
import hashlib
from io import BytesIO
from logging import getLogger

import OpenSSL.crypto
import asn1crypto.ocsp
import pytz
from OpenSSL import crypto
from OpenSSL.crypto import X509StoreContextError
from asn1crypto import pem

from bankid.experimental.helper import CompletionDataContainer, make_cert, NonceParse

_LOG = getLogger(__name__)


def verify_bankid_response(bank_id_response, ensure_certificates_still_valid=True, BANK_ID_ROOT_CERT=None):

    if not isinstance(bank_id_response, dict):
        raise TypeError("Response not a dictionary")

    if "completionData" not in bank_id_response:
        raise AttributeError("Completion data missing in dictionary")

    try:
        cdc = CompletionDataContainer(bank_id_response["completionData"])

        # First step is to hash the data and verify the digest matches

        _LOG.info("1. Message Digest Verification\n")
        bid_signed_data_raw_bytes = cdc.signature_container.bid_signed_data_raw.encode()

        # TODO - Parse out of the XML which hashing algorithm should be sued

        bid_signed_data_hash = hashlib.sha256(bid_signed_data_raw_bytes).digest().hex()
        key_info_hash = hashlib.sha256(cdc.signature_container.key_info_raw.encode()).digest().hex()

        signed_data_hash_from_signature = base64.b64decode(cdc.signature_container.signed_data_digest.text).hex()
        key_info_hash_from_signature = base64.b64decode(cdc.signature_container.key_data_digest.text).hex()

        if bid_signed_data_hash != signed_data_hash_from_signature:
            raise AssertionError("Signed Data hash does not match!")

        if key_info_hash != key_info_hash_from_signature:
            raise AssertionError("Key Info hash does not match!")

        _LOG.info("\n2. Signature verification\n")

        # Helper function for the certificates
        user_certificate_string = make_cert(cdc.signature_container.certificates[0].text)

        # Making a certificate object out of it
        user_certificate = crypto.load_certificate(
            crypto.FILETYPE_PEM, BytesIO(user_certificate_string.encode()).read()
        )

        signature_bytes = base64.b64decode(cdc.signature_container.signature_value.text)
        signed_info = cdc.signature_container.signed_info.encode()

        try:
            _LOG.debug("Certificate:", user_certificate.get_subject())
            _LOG.debug("Signature Bytes:", signature_bytes)
            _LOG.debug("Signature Data Raw:", signed_info)

            OpenSSL.crypto.verify(user_certificate, signature_bytes, signed_info, "sha256")
        except OpenSSL.crypto.Error as e:
            raise AssertionError("The BankID signature is not valid!")

        _LOG.info("\n3. OCSP Response Verification\n")

        ocsp = base64.b64decode(bank_id_response["completionData"]["ocspResponse"])
        ocsp_response = asn1crypto.ocsp.OCSPResponse.load(ocsp)

        basic_ocsp_response = ocsp_response["response_bytes"]["response"].parsed

        # Some help by listing all the different parts of the OCSP response
        _LOG.debug("TBS Response Data", basic_ocsp_response["tbs_response_data"])
        _LOG.debug("SignatureAlgorithm", basic_ocsp_response["signature_algorithm"].signature_algo)
        _LOG.debug("SignatureAlgorithm Hash Function", basic_ocsp_response["signature_algorithm"].hash_algo)
        _LOG.debug("Signature", basic_ocsp_response["signature"].__bytes__())
        _LOG.debug("Cert", basic_ocsp_response["certs"])

        # Response content
        _LOG.debug("version", basic_ocsp_response["tbs_response_data"]["version"])
        _LOG.debug("responderID", basic_ocsp_response["tbs_response_data"]["responder_id"])  # has native
        _LOG.info("producedAt", basic_ocsp_response["tbs_response_data"]["produced_at"])

        cest = pytz.timezone("Europe/Stockholm")
        ocsp_produced_at = basic_ocsp_response["tbs_response_data"]["produced_at"].native

        if not isinstance(ocsp_produced_at, datetime.datetime):
            raise AssertionError("OCSP produced at is not a datetime!")

        ocsp_produced_at = ocsp_produced_at.astimezone(cest).strftime("%Y-%m-%d %H:%M:%S")

        _LOG.debug("responses", basic_ocsp_response["tbs_response_data"]["responses"])
        _LOG.debug("response Extentions", basic_ocsp_response["tbs_response_data"]["response_extensions"])

        _LOG.debug("Extentions")
        extention = basic_ocsp_response["tbs_response_data"]["response_extensions"][0]

        _LOG.debug("extn_id", extention["extn_id"])
        _LOG.debug("critical", extention["critical"])

        # Cannot _LOG.debug the value without an exception being raised - need to parse that ourself later
        # print ('extn_value', extention['extn_value'])

        single_response = basic_ocsp_response["tbs_response_data"]["responses"][0]

        _LOG.debug("CertID", single_response["cert_id"])
        _LOG.debug("certStatus", single_response["cert_status"])
        _LOG.debug("thisUpdate", single_response["this_update"])
        _LOG.debug("nextUpdate", single_response["next_update"])
        _LOG.debug("singleExtensions", single_response["single_extensions"])

        _LOG.info("3.1. OCSP Response - Verify success ")

        if ocsp_response["response_status"].native != "successful":
            raise AssertionError("OCSP response status was not successful")

        _LOG.info("3.2. OCSP Response - Verify signature ")

        # Transform the asn1 certificate to an openssl certificate
        der_bytes = basic_ocsp_response["certs"][0].dump()
        pem_bytes = pem.armor("CERTIFICATE", der_bytes)
        ocsp_certificate = crypto.load_certificate(crypto.FILETYPE_PEM, pem_bytes)

        # Get the signature bytes
        signature = basic_ocsp_response["signature"].__bytes__()

        # Dump the TBS response data as DER bytes
        signature_data = basic_ocsp_response["tbs_response_data"].dump()

        # Define the hashing algorithm to be used
        digest_method = basic_ocsp_response["signature_algorithm"].hash_algo

        _LOG.debug("Certificate", ocsp_certificate.get_subject())
        _LOG.debug("Signature", signature)
        _LOG.debug("Signature data", signature_data)
        _LOG.debug("Digest Method", digest_method)

        try:
            OpenSSL.crypto.verify(ocsp_certificate, signature, signature_data, digest_method)
        except OpenSSL.crypto.Error as e:
            raise AssertionError("The OCSP signature is not valid!")

        _LOG.info("3.2. OCSP Response - Compare nonce")

        nonce_computed = hashlib.sha1(bank_id_response["completionData"]["signature"].encode("utf-8")).digest().hex()

        # A helper because the asn1 library seems to have a problem with the nonce parsing in some form or the other
        nonce_parser = NonceParse(extention.contents)

        # Verify that the computed nonce is part of the nonce value given in the oscp
        # Note that it only partially matches as we use sha-1 to compute the hash

        _LOG.debug("Nonce value computed ", nonce_computed)
        _LOG.debug("Nonce value presented", nonce_parser.value.hex())

        if not nonce_parser.value.hex().startswith(nonce_computed):
            raise AssertionError("Computed nonce not matching the OCSP nonce")

        _LOG.info("\n4. Verify all the certificates by relying on the BankID root certificate as a trusted one \n")

        user_cert = crypto.load_certificate(
            crypto.FILETYPE_PEM, make_cert(cdc.signature_container.certificates[0].text).encode()
        )

        bank_user_cert = crypto.load_certificate(
            crypto.FILETYPE_PEM, make_cert(cdc.signature_container.certificates[1].text).encode()
        )

        bank_bank_id_cert = crypto.load_certificate(
            crypto.FILETYPE_PEM, make_cert(cdc.signature_container.certificates[2].text).encode()
        )

        bank_id_root_cert = crypto.load_certificate(crypto.FILETYPE_PEM, BANK_ID_ROOT_CERT.encode())

        # 3. verify the certificate chain of the tbs certificate

        # Make sure we respect or do not respect certificate expiration times
        if not ensure_certificates_still_valid:
            tomorrow = datetime.datetime.now() + datetime.timedelta(days=1)

            bank_user_cert.set_notAfter(tomorrow.strftime("%Y%m%d%H%M%SZ").encode())
            bank_bank_id_cert.set_notAfter(tomorrow.strftime("%Y%m%d%H%M%SZ").encode())
            bank_id_root_cert.set_notAfter(tomorrow.strftime("%Y%m%d%H%M%SZ").encode())

            ocsp_certificate.set_notAfter(tomorrow.strftime("%Y%m%d%H%M%SZ").encode())
            user_cert.set_notAfter(tomorrow.strftime("%Y%m%d%H%M%SZ").encode())

        store = crypto.X509Store()
        store.add_cert(bank_user_cert)
        store.add_cert(bank_bank_id_cert)
        store.add_cert(bank_id_root_cert)

        try:
            # Verify the user certificate up to the root certificate
            store_ctx = crypto.X509StoreContext(store, user_cert)
            store_ctx.verify_certificate()
            _LOG.debug("User Certificate issued by the respective bank... OK")
        except X509StoreContextError:
            raise AssertionError("BankID user certificate chain could not be verified.")

        try:
            # Verify the ocsp certificate up to the root certificate
            store_ctx = crypto.X509StoreContext(store, ocsp_certificate)
            store_ctx.verify_certificate()
            _LOG.debug("OCSP Certificate issued by the respective bank... OK")
        except X509StoreContextError:
            raise AssertionError("OCSP certificate chain could not be verified.")

    except Exception as e:
        raise e

    return ocsp_produced_at



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`bankid.jsonclient` -- BankID JSON Client
==============================================

Created on 2018-02-19 by hbldh

"""
import base64
from urllib import parse as urlparse

import requests
from pkg_resources import resource_filename

from bankid.exceptions import get_json_error_class


def _encode_user_data(user_data):
    if isinstance(user_data, str):
        return base64.b64encode(user_data.encode("utf-8")).decode("ascii")
    else:
        return base64.b64encode(user_data).decode("ascii")


class BankIDJSONClient(object):
    """The client to use for communicating with BankID servers via the v.5 API.

    :param certificates: Tuple of string paths to the certificate to use and
        the key to sign with.
    :type certificates: tuple
    :param test_server: Use the test server for authenticating and signing.
    :type test_server: bool
    :param request_timeout: Timeout for BankID requests.
    :type request_timeout: int

    """

    def __init__(self, certificates, test_server=False, request_timeout=None):
        self.certs = certificates
        self._request_timeout = request_timeout

        if test_server:
            self.api_url = "https://appapi2.test.bankid.com/rp/v5.1/"
            self.verify_cert = resource_filename("bankid.certs", "appapi2.test.bankid.com.pem")
        else:
            self.api_url = "https://appapi2.bankid.com/rp/v5.1/"
            self.verify_cert = resource_filename("bankid.certs", "appapi2.bankid.com.pem")

        self.client = requests.Session()
        self.client.verify = self.verify_cert
        self.client.cert = self.certs
        self.client.headers = {"Content-Type": "application/json"}

        self._auth_endpoint = urlparse.urljoin(self.api_url, "auth")
        self._sign_endpoint = urlparse.urljoin(self.api_url, "sign")
        self._collect_endpoint = urlparse.urljoin(self.api_url, "collect")
        self._cancel_endpoint = urlparse.urljoin(self.api_url, "cancel")

    def _post(self, endpoint, *args, **kwargs):
        """Internal helper method for adding timeout to requests."""
        return self.client.post(endpoint, *args, timeout=self._request_timeout, **kwargs)

    def authenticate(self, end_user_ip, personal_number=None, requirement=None, **kwargs):
        """Request an authentication order. The :py:meth:`collect` method
        is used to query the status of the order.

        Note that personal number is not needed when authentication is to
        be done on the same device, provided that the returned
        ``autoStartToken`` is used to open the BankID Client.

        Example data returned:

        .. code-block:: json

            {
                'orderRef': 'ee3421ea-2096-4000-8130-82648efe0927',
                'autoStartToken': 'e8df5c3c-c67b-4a01-bfe5-fefeab760beb',
                'qrStartToken': '01f94e28-857f-4d8a-bf8e-6c5a24466658',
                'qrStartSecret': 'b4214886-3b5b-46ab-bc08-6862fddc0e06'
            }

        :param end_user_ip: IP address of the user requesting
            the authentication.
        :type end_user_ip: str
        :param personal_number: The Swedish personal number in
            format YYYYMMDDXXXX.
        :type personal_number: str
        :param requirement: An optional dictionary stating how the signature
            must be created and verified. See BankID Relying Party Guidelines,
            section 13.5 for more details.
        :type requirement: dict
        :return: The order response.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                             when error has been returned from server.

        """
        data = {"endUserIp": end_user_ip}
        if personal_number:
            data["personalNumber"] = personal_number
        if requirement and isinstance(requirement, dict):
            data["requirement"] = requirement
        # Handling potentially changed optional in-parameters.
        data.update(kwargs)
        response = self._post(self._auth_endpoint, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise get_json_error_class(response)

    def sign(
        self,
        end_user_ip,
        user_visible_data,
        personal_number=None,
        requirement=None,
        user_non_visible_data=None,
        **kwargs
    ):
        """Request a signing order. The :py:meth:`collect` method
        is used to query the status of the order.

        Note that personal number is not needed when signing is to be done
        on the same device, provided that the returned ``autoStartToken``
        is used to open the BankID Client.

        Example data returned:

        .. code-block:: json

            {
                'orderRef': 'ee3421ea-2096-4000-8130-82648efe0927',
                'autoStartToken': 'e8df5c3c-c67b-4a01-bfe5-fefeab760beb',
                'qrStartToken': '01f94e28-857f-4d8a-bf8e-6c5a24466658',
                'qrStartSecret': 'b4214886-3b5b-46ab-bc08-6862fddc0e06'
            }

        :param end_user_ip: IP address of the user requesting
            the authentication.
        :type end_user_ip: str
        :param user_visible_data: The information that the end user
            is requested to sign.
        :type user_visible_data: str
        :param personal_number: The Swedish personal number in
            format YYYYMMDDXXXX.
        :type personal_number: str
        :param requirement: An optional dictionary stating how the signature
            must be created and verified. See BankID Relying Party Guidelines,
            section 13.5 for more details.
        :type requirement: dict
        :param user_non_visible_data: Optional information sent with request
            that the user never sees.
        :type user_non_visible_data: str
        :return: The order response.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                     when error has been returned from server.

        """
        data = {"endUserIp": end_user_ip}
        if personal_number:
            data["personalNumber"] = personal_number
        data["userVisibleData"] = _encode_user_data(user_visible_data)
        if user_non_visible_data:
            data["userNonVisibleData"] = _encode_user_data(user_non_visible_data)
        if requirement and isinstance(requirement, dict):
            data["requirement"] = requirement
        # Handling potentially changed optional in-parameters.
        data.update(kwargs)
        response = self._post(self._sign_endpoint, json=data)

        if response.status_code == 200:
            return response.json()
        else:
            raise get_json_error_class(response)

    def collect(self, order_ref):
        """Collects the result of a sign or auth order using the
        ``orderRef`` as reference.

        RP should keep on calling collect every two seconds as long as status
        indicates pending. RP must abort if status indicates failed. The user
        identity is returned when complete.

        Example collect results returned while authentication or signing is
        still pending:

        .. code-block:: json

            {
                "orderRef":"131daac9-16c6-4618-beb0-365768f37288",
                "status":"pending",
                "hintCode":"userSign"
            }

        Example collect result when authentication or signing has failed:

        .. code-block:: json

            {
                "orderRef":"131daac9-16c6-4618-beb0-365768f37288",
                "status":"failed",
                "hintCode":"userCancel"
            }

        Example collect result when authentication or signing is successful
        and completed:

        .. code-block:: json

            {
                "orderRef":"131daac9-16c6-4618-beb0-365768f37288",
                "status":"complete",
                "completionData": {
                    "user": {
                        "personalNumber":"190000000000",
                        "name":"Karl Karlsson",
                        "givenName":"Karl",
                        "surname":"Karlsson"
                    },
                    "device": {
                        "ipAddress":"192.168.0.1"
                    },
                    "cert": {
                        "notBefore":"1502983274000",
                        "notAfter":"1563549674000"
                    },
                    "signature":"<base64-encoded data>",
                    "ocspResponse":"<base64-encoded data>"
                }
            }

        See `BankID Relying Party Guidelines Version: 3.5 <https://www.bankid.com/assets/bankid/rp/bankid-relying-party-guidelines-v3.5.pdf>`_
        for more details about how to inform end user of the current status,
        whether it is pending, failed or completed.

        :param order_ref: The ``orderRef`` UUID returned from auth or sign.
        :type order_ref: str
        :return: The CollectResponse parsed to a dictionary.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                             when error has been returned from server.

        """
        response = self._post(self._collect_endpoint, json={"orderRef": order_ref})

        if response.status_code == 200:
            return response.json()
        else:
            raise get_json_error_class(response)

    def cancel(self, order_ref):
        """Cancels an ongoing sign or auth order.

        This is typically used if the user cancels the order
        in your service or app.

        :param order_ref: The UUID string specifying which order to cancel.
        :type order_ref: str
        :return: Boolean regarding success of cancellation.
        :rtype: bool
        :raises BankIDError: raises a subclass of this error
                             when error has been returned from server.

        """
        response = self._post(self._cancel_endpoint, json={"orderRef": order_ref})

        if response.status_code == 200:
            return response.json() == {}
        else:
            raise get_json_error_class(response)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

import bankid


# def test_certutils_main():
#     bankid.certutils.main(verbose=False)
#
#     assert os.path.exists(os.path.expanduser("~/certificate.pem"))
#     assert os.path.exists(os.path.expanduser("~/key.pem"))
#
#     try:
#         os.remove(os.path.expanduser("~/certificate.pem"))
#         os.remove(os.path.expanduser("~/key.pem"))
#     except:
#         pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`bankid`
=============

Created 2016-11-28 by hbldh

PyBankID is a client for providing BankID services as a Relying Party, i.e.
providing authentication and signing functionality to end users. This package
provides a simplifying interface for initiating authentication
and signing orders and then collecting the results from the BankID servers.

If you intend to use PyBankID in your project, you are advised to read
the `BankID Relying Party Guidelines
<https://www.bankid.com/utvecklare/rp-info>`_ before
doing anything else. There, one can find information
about how the BankID methods are defined and how to use them.

"""

from .jsonclient import BankIDJSONClient
from .certutils import create_bankid_test_server_cert_and_key
from .__version__ import __version__, version
import bankid.exceptions

__all__ = [
    "BankIDJSONClient",
    "exceptions",
    "create_bankid_test_server_cert_and_key",
    "__version__",
    "version",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`test_client`
==================

.. module:: test_client
   :platform: Unix, Windows
   :synopsis:

.. moduleauthor:: hbldh <henrik.blidh@nedomkull.com>

Created on 2015-08-07, 12:00

"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

import random
import tempfile
import uuid

import pytest

try:
    from unittest import mock
except:
    import mock

import bankid


def _get_random_personal_number():
    """Simple random Swedish personal number generator."""

    def _luhn_digit(id_):
        """Calculate Luhn control digit for personal number.

        Code adapted from `Faker
        <https://github.com/joke2k/faker/blob/master/faker/providers/ssn
        /sv_SE/__init__.py>`_.

        :param id_: The partial number to calculate checksum of.
        :type id_: str
        :return: Integer digit in [0, 9].
        :rtype: int

        """

        def digits_of(n):
            return [int(i) for i in str(n)]

        id_ = int(id_) * 10
        digits = digits_of(id_)
        checksum = sum(digits[-1::-2])
        for k in digits[-2::-2]:
            checksum += sum(digits_of(k * 2))
        checksum %= 10

        return checksum if checksum == 0 else 10 - checksum

    year = random.randint(1900, 2014)
    month = random.randint(1, 12)
    day = random.randint(1, 28)
    suffix = random.randint(0, 999)
    pn = "{0:04d}{1:02d}{2:02d}{3:03d}".format(year, month, day, suffix)
    return pn + str(_luhn_digit(pn[2:]))


def test_authentication_and_collect(cert_and_key, ip_address):
    """Authenticate call and then collect with the returned orderRef UUID."""

    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=True)
    assert "appapi2.test.bankid.com.pem" in c.verify_cert
    out = c.authenticate(ip_address, _get_random_personal_number())
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out.get("orderRef"), version=4)
    collect_status = c.collect(out.get("orderRef"))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")


def test_sign_and_collect(cert_and_key, ip_address):
    """Sign call and then collect with the returned orderRef UUID."""

    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=True)
    out = c.sign(
        ip_address,
        "The data to be signed",
        personal_number=_get_random_personal_number(),
        user_non_visible_data="Non visible data",
    )
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out.get("orderRef"), version=4)
    collect_status = c.collect(out.get("orderRef"))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")


def test_invalid_orderref_raises_error(cert_and_key):
    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=True)
    with pytest.raises(bankid.exceptions.InvalidParametersError):
        collect_status = c.collect("invalid-uuid")


def test_already_in_progress_raises_error(cert_and_key, ip_address):
    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=True)
    pn = _get_random_personal_number()
    out = c.authenticate(ip_address, pn)
    with pytest.raises(bankid.exceptions.AlreadyInProgressError):
        out2 = c.authenticate(ip_address, pn)


def test_already_in_progress_raises_error_2(cert_and_key, ip_address):
    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=True)
    pn = _get_random_personal_number()
    out = c.sign(ip_address, "Text to sign", pn)
    with pytest.raises(bankid.exceptions.AlreadyInProgressError):
        out2 = c.sign(ip_address, "Text to sign", pn)


def test_authentication_and_cancel(cert_and_key, ip_address):
    """Authenticate call and then cancel it"""

    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=True)
    out = c.authenticate(ip_address, _get_random_personal_number())
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out.get("orderRef"), version=4)
    collect_status = c.collect(out.get("orderRef"))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")
    success = c.cancel(str(order_ref))
    assert success
    with pytest.raises(bankid.exceptions.InvalidParametersError):
        collect_status = c.collect(out.get("orderRef"))


def test_cancel_with_invalid_uuid(cert_and_key):
    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=True)
    invalid_order_ref = uuid.uuid4()
    with pytest.raises(bankid.exceptions.InvalidParametersError):
        cancel_status = c.cancel(str(invalid_order_ref))


@pytest.mark.parametrize(
    "test_server, endpoint",
    [(False, "appapi2.bankid.com"), (True, "appapi2.test.bankid.com")],
)
def test_correct_prod_server_urls(cert_and_key, test_server, endpoint):
    c = bankid.BankIDJSONClient(certificates=cert_and_key, test_server=test_server)
    assert c.api_url == "https://{0}/rp/v5.1/".format(endpoint)
    assert "{0}.pem".format(endpoint) in c.verify_cert



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
PyBankID
========

PyBankID is a client for performing BankID signing.

The Swedish BankID solution for digital signing utilizes a
JSON API, and this module aims at providing a simplifying
client for making authentication, signing and collect requests to
the BankID servers.

The latest development version is available at the project's `GitHub
site <https://github.com/hbldh/pybankid/>`_.

Created by hbldh <henrik.blidh@nedomkull.com>
Created on 2013-09-14, 19:31

"""

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "pybankid"
DESCRIPTION = "BankID Relying Party client for Python"
URL = "https://github.com/hbldh/pybankid"
EMAIL = "henrik.blidh@nedomkull.com"
AUTHOR = "Henrik Blidh"

# What packages are required for this module to be executed?
REQUIRED = ["requests", "six"]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, "bankid", "__version__.py")) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=("tests",)),
    install_requires=REQUIRED,
    include_package_data=True,
    package_data={"": ["*.pem", "*.p12"]},
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Security",
        "Topic :: Utilities",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
    extras_require={
        "security": ["pyOpenSSL>=0.13", "ndg-httpsclient", "pyasn1"],
        "signature-verification": {"pyOpenSSL", "asn1crypto", "freezegun", "pytz"},
    },
)



import base64
import xml.etree.ElementTree as ET
from textwrap import wrap

make_cert = lambda e: "-----BEGIN CERTIFICATE-----\n" + "\n".join(wrap(e, 54)) + "\n-----END CERTIFICATE-----"


class B64Value:
    def __init__(self, value):
        self.value = value

    @property
    def decode(self):
        return base64.b64decode(self.value)

    def raw(self):
        return self.value

    def __str__(self):
        return self.raw()


class BankIdSignatureContainer:
    def __init__(self, signature):
        self.root = ET.fromstring(signature.decode)
        self.raw = signature

    @property
    def signature_value(self):
        return B64Value(self.root[1].text).decode

    @property
    def signed_data_digest(self):
        return self.root[0][2][2]

    @property
    def key_data_digest(self):
        return self.root[0][3][2]

    @property
    def signature_value(self):
        return self.root[1]

    @property
    def signed_data_raw(self):
        return self.root[3][0]

    @property
    def certificates(self):
        return [e for e in self.root[2][0]]

    @property
    def bid_signed_data_raw(self):
        raw_xml_string = self.raw.decode.decode()
        start = raw_xml_string.find("<Object>") + len("<Object>")
        stop = raw_xml_string.find("</Object>")
        return raw_xml_string[start:stop]

    @property
    def user_non_visible_data(self):
        return self.root[3][0][1].text

    @property
    def signed_info(self):
        raw_xml_string = self.raw.decode.decode()
        start = raw_xml_string.find("<SignedInfo")
        stop = raw_xml_string.find("</SignedInfo>")
        return raw_xml_string[start:stop] + "</SignedInfo>"

    @property
    def key_info_raw(self):
        raw_xml_string = self.raw.decode.decode()
        start = raw_xml_string.find("<KeyInfo")
        stop = raw_xml_string.find("</KeyInfo>")
        return raw_xml_string[start:stop] + "</KeyInfo>"

    @property
    def server_info(self):
        return {
            "name": B64Value(self.root[3][0][2][0].text).decode,
            "displayName": B64Value(self.root[3][0][2][2].text).decode,
        }


class CompletionDataContainer:
    def __init__(self, completion_data):
        self.completion_data = completion_data

    @property
    def order_ref(self):
        return self.completion_data["orderRef"]

    @property
    def ocsp_response(self):
        return self.completion_data["ocspResponse"]

    @property
    def device(self):
        return self.completion_data["device"]

    @property
    def user(self):
        return self.completion_data["user"]

    @property
    def signature(self):
        return B64Value(self.completion_data["signature"])

    @property
    def signature_container(self):
        return BankIdSignatureContainer(self.signature)


class NonceParse:
    def __init__(self, bytes):
        self.bytes = bytes

    @property
    def type(self):
        return ".".join(str(e) for e in list(self.bytes.contents[2:11]))

    @property
    def critical(self):
        return self.bytes.contents[13] == 255

    @property
    def value(self):
        return self.bytes[16:]
